"""
Utilities for working with files
"""
import json
import os
from typing import Dict, List, Optional

import jsonschema
import yaml

from core_library.utilities.misc_utils import setup_console_logger

mds_logger = setup_console_logger()


def yaml_read_file(file_path: str) -> Dict:
    """
    Safely loads a yaml file into a dictionary

    :param file_path: path to the file
    :type file_path: str
    :return: yaml file as a dictionary
    :rtype: Dict
    """
    mds_logger.info(f"Loading YAML file: {file_path}")
    with open(file_path, "r") as file:
        data = yaml.safe_load(file)
    return data


def yaml_validate_schema(yaml: Dict, schema: Dict) -> bool:
    """
    YAML Validate the file with a predefined JSON schema

    :param yaml: YAML data
    :type yaml: Dict
    :param schema: JSON Schema
    :type schema: Dict
    :raises e: Validation Error
    :return: True if pass
    :rtype: bool
    """
    try:
        jsonschema.validate(instance=yaml, schema=schema)
        mds_logger.info("YAML file passed validation")
    except jsonschema.ValidationError as e:
        mds_logger.error("YAML file failed validation")
        raise e
    else:
        return True


def get_files_directory(
    directory: str, file_extension: Optional[str] = None
) -> List[str]:
    """
    Get file paths from a directory recursively

    :param directory: path to check
    :type directory: str
    :param file_extension: file_extension to include
    :type file_extension: Optional[str]
    :return: List of file paths
    :rtype: List[str]
    """
    file_paths = []
    for root, dir, files in os.walk(directory):
        for file in files:
            if file_extension:
                if file.endswith(file_extension):
                    file_path = os.path.join(root, file)
                    file_paths.append(file_path)
            else:
                file_path = os.path.join(root, file)
                file_paths.append(file_path)

    return file_paths


def json_read_file(file_path: str) -> Dict:
    """
    Reads a json file and returns the data in a dict

    :param file_path: file path
    :type file_path: str
    :return: data
    :rtype: Dict
    """
    with open(file_path) as f:
        data = json.load(f)
    return data



"""
Python module that loads YAML config to Dagster Asset Checks from YAML
"""
from enum import Enum
from typing import Generator, List

import polars as pl
from dagster import (
    AssetCheckResult,
    AssetChecksDefinition,
    AssetExecutionContext,
    asset_check,
)
from jinja2 import Template
from jsonschema.exceptions import ValidationError

from core_library.utilities.data_utils import key_values_in_lod
from core_library.utilities.file_utils import (
    get_files_directory,
    json_read_file,
    yaml_read_file,
    yaml_validate_schema,
)
from core_library.utilities.misc_utils import setup_console_logger

mds_logger = setup_console_logger()


class Engines(Enum):
    """
    Engines we have defined for Asset Checks
    """

    polars = "polars"


class PolarsAssetChecks:
    """
    Asset checks for polars
    Each check should return the 'illegal' records
    """

    def main_handler(self, test_name: str, column: str) -> str:
        """
        Function that handles what Data Asset Check to call

        :param test_name: test name to run
        :type test_name: str
        :param column: column to run the test on
        :type column: str
        :raises Exception: If check is not implemented
        :return: code to execute
        :rtype: str
        """
        if test_name == "not_null":
            return self.not_null(column=column)
        elif test_name == "unique":
            return self.unique(column=column)
        else:
            raise Exception(f"Asset Check {test_name} is not configured")

    def not_null(self, column: str) -> str:
        """
        Polars Data Test for not_null columns

        :param column: column to check
        :type column: str
        :return: Failed records
        :rtype: str
        """
        template = Template("df.filter(pl.col('{{ column }}').is_null())")
        generated_code = template.render(column=column)
        return generated_code

    def unique(self, column: str) -> str:
        """
        Polars Data Test for unique columns

        :param column: column to check
        :type column: str
        :return: Failed records
        :rtype: str
        """
        template = Template("df.filter(pl.col('{{ column }}').is_duplicated())")
        generated_code = template.render(column=column)
        return generated_code


def load_yaml_asset_check_files(
    directory: str = "mds_dagster/asset_checks"
) -> Generator[List, None, None]:
    """
    Load all the YAML Asset Check files into a Generator of lists

    :param directory: directory to load yaml files from, defaults to "mds_dagster/asset_checks"
    :type directory: str, optional
    :yield: List data from YAML file of tests
    :rtype: Generator
    """
    files = get_files_directory(directory=directory, file_extension=".yaml")
    for file in files:
        asset_check_def = yaml_read_file(file_path=file)
        # Validate the schema
        try:
            asset_check_schema = json_read_file(
                "mds_dagster/asset_checks/asset_checks_schema.json"
            )
            yaml_validate_schema(asset_check_def, asset_check_schema)
        except ValidationError:
            mds_logger.error(f"File: {file} failed yaml asset_check_schema validation")
            continue

        # Validate the engine value
        engines = key_values_in_lod(
            asset_check_def.get("data_tests"), select_key="engine"
        )
        for engine in engines:
            assert hasattr(
                Engines, engine
            ), f"Asset Check Factory Error - File {file} has invalid engine."

        # TODO: We need to validate all of the test_names that they are valid test names

        data = asset_check_def.get("data_tests")
        assert isinstance(data, List)

        yield data


def create_dagster_asset_check_name(
    asset_name: str, check_name: str, column_name: str
) -> str:
    """
    Create the Dagster Asset Check Name. What dagster registers the asset check as

    :param asset_name: asset name to run the check on
    :type asset_name: str
    :param check_name: the check to run
    :type check_name: str
    :param column_name: the columnn to run the check on
    :type column_name: str
    :return: asset check name
    :rtype: str
    """

    check_name = "asset_check__" + asset_name + "_" + check_name + "_" + column_name

    return check_name


def create_dagster_check(
    asset_name: str, check_name: str, column_name: str, engine_name: str
) -> AssetChecksDefinition:
    """
    Function that creates the dagster asset check
    :param asset_name: asset name to run the check on
    :type asset_name: str
    :param check_name: the check to run
    :type check_name: str
    :param column_name: the columnn to run the check on
    :type column_name: str
    :param engine_name: the engine to use
    :type engine_name: str
    :raises Exception: If engine is not configured
    :return: Dagster Asset Check
    :rtype: AssetChecksDefinition
    """
    polars_asset_checks = PolarsAssetChecks()
    asset_check_name = create_dagster_asset_check_name(
        asset_name, check_name, column_name
    )

    @asset_check(
        name=asset_check_name,
        asset=asset_name,
        required_resource_keys={
            "strava_api_resource",
            "polars_parquet_io_manager_strava_ingest",
        },
    )
    def _check(context: AssetExecutionContext) -> AssetCheckResult:
        """
        Helper function for registering and executing the check

        :param context: Dagster context
        :type context: AssetExecutionContext
        :raises Exception: Exception if engine is not configured
        :return: Dagster Asset Check Result
        :rtype: AssetCheckResult
        """
        if engine_name == "polars":
            # TODO: This is not currently very flexible as it is assuming a base directory
            # And it is tied to the Strava Ingest
            base_directory = (
                context.resources.polars_parquet_io_manager_strava_ingest.base_dir
            )

            mds_logger.info(f"Setting Directory for parquet - {base_directory}")
            df = pl.read_parquet(source=f"{base_directory}/{asset_name}.parquet")  # noqa: F841
            code_to_run = polars_asset_checks.main_handler(
                test_name=check_name, column=column_name
            )

            result_df = eval(code_to_run)
            assert isinstance(result_df, pl.DataFrame)

            return AssetCheckResult(passed=len(result_df) == 0)
        else:
            raise Exception(f"Asset Check Engine {engine_name} is not configured")

    return _check


def dagster_load_all_checks() -> List[AssetChecksDefinition]:
    """
    Function that loads all the Dagster Asset Checks into a list

    :return: Dagster Asset Checks
    :rtype: List[AssetChecksDefinition]
    """
    check_results = []
    checks = load_yaml_asset_check_files()
    # Parse the asset check dictionary
    for assets in checks:
        for asset in assets:
            asset_name = asset["asset"]
            engine_name = asset["engine"]
            for test in asset.get("tests"):
                test_name = test.get("name")
                for column in test.get("columns"):
                    result = create_dagster_check(
                        asset_name=asset_name,
                        check_name=test_name,
                        column_name=column,
                        engine_name=engine_name,
                    )
                    check_results.append(result)
    return check_results



from core_library.dagster.dagster_asset_check_factory import (
    PolarsAssetChecks,
    create_dagster_check,
    dagster_load_all_checks,
    load_yaml_asset_check_files,
)
from core_library.handler.strava_api import StravaHandler as StravaHandler
from core_library.utilities.data_utils import key_values_in_lod
from core_library.utilities.date_utils import (
    get_current_epoch_time,
)
from core_library.utilities.file_utils import (
    get_files_directory,
    json_read_file,
    yaml_read_file,
    yaml_validate_schema,
)
from core_library.utilities.misc_utils import setup_console_logger
from core_library.utilities.polars_dataframe_utils import (
    pl_create_df,
    pl_df_cols_to_standard,
)
from core_library.utilities.text_utils import (
    camel_case_to_snake_case,
    cols_text_to_standard,
    remove_special_charachters,
)

__all__ = [
    "StravaHandler",
    "setup_console_logger",
    "get_current_epoch_time",
    "key_values_in_lod",
    "get_files_directory",
    "json_read_file",
    "yaml_read_file",
    "yaml_validate_schema",
    "pl_create_df",
    "pl_df_cols_to_standard",
    "camel_case_to_snake_case",
    "cols_text_to_standard",
    "remove_special_charachters",
    "PolarsAssetChecks",
    "load_yaml_asset_check_files",
    "create_dagster_check",
    "dagster_load_all_checks",
]



"""
Strava API Handler
"""
from functools import lru_cache
from typing import Dict, Generator, List, Optional

import requests

# from tenacity import retry, stop_after_attempt, wait_fixed
from core_library.utilities.misc_utils import setup_console_logger

mds_logger = setup_console_logger()


class StravaHandler:
    """
    Strava Handler for communicating with API
    """

    def __init__(
        self,
        strava_client_id: str,
        strava_client_secret: str,
        grant_type: str,
        code: Optional[str] = None,
        refresh_token: Optional[str] = None,
    ):
        """Init class for the API

        Args:
            strava_client_id (str): Strava API Client ID
            strava_client_secret (str): Strava API Client Secret
            code (Optional[str], optional): Strava API Code for authorization_code
            refresh_token (Optional[str], optional): Refresh Token if using the refresh_token grant_type
            grant_type (Optional[str], optional): _description_. Defaults to "authorization_code".
        """
        mds_logger.info("Initiallizing the Strava client")
        self.strava_client_id = strava_client_id
        self.strava_client_secret = strava_client_secret
        self.code = code
        self.grant_type = grant_type
        self.refresh_token = refresh_token
        self.base_url = "https://www.strava.com/api/v3/"

    # TODO: implement a _post method for getting the data
    @lru_cache(maxsize=100)
    # TODO: I have to fix this for pytest (retry)
    # @retry(stop=stop_after_attempt(3), wait=wait_fixed(3))
    def generate_token(self) -> dict:
        """
        Gets a Stava API Bearer Token and associated data

        Returns:
            dict: _description_
        """

        mds_logger.info("Generating a Strava API Token")

        if self.grant_type == "authorization_code" and self.code is not None:
            # Construct URL
            url = (
                f"{self.base_url}oauth/token?client_id={self.strava_client_id}"
                f"&client_secret={self.strava_client_secret}"
                f"&grant_type={self.grant_type}&code={self.code}"
            )
        # refresh_token
        elif self.grant_type == "refresh_token" and self.refresh_token is not None:
            # Construct URL
            url = (
                f"{self.base_url}oauth/token?client_id={self.strava_client_id}"
                f"&client_secret={self.strava_client_secret}"
                f"&grant_type={self.grant_type}&refresh_token={self.refresh_token}"
            )

        else:
            raise Exception(
                "Please pass in either (authorization_code and code) or (refresh_token and refresh_token)"
            )
        # Make the Call
        response = requests.post(
            url=url,
            headers={},
            data={},
        )
        # Call Failed
        if response.status_code != 200:
            raise Exception(
                "Failed to generate Stava Token \n"
                f"status code: {response.status_code} \n"
                f"response: {response.text}"
            )
        # Parse response
        response = response.json()

        if "access_token" in response:
            return response.get("access_token")
        raise Exception("No token found in response")

    def get_api_headers(self) -> dict:
        """
        Returns the API Headers needed for the call
        """
        return {"Authorization": f"Bearer {self.generate_token()}"}

    def get_athlete(self) -> List:
        """
        Get Athelete Data - Includes basic information on athlete

        :yield: Data from API Call
        :rtype: Generator[dict, None, None]
        """

        mds_logger.info("Fetching Athlete Data")
        data = self._get(endpoint="athlete")

        return list(data)

    def get_equipment(self, ids: List[str]) -> List:
        """
        Equipment Data of stats on the equipment.

        :param ids: List of equipemnt id from strava
        :type ids: List[str]
        :return: data
        :rtype: dict
        """
        lod = []
        for id in ids:
            mds_logger.info(f"Fetching Equpment Data - {id}")
            data = self._get(endpoint=f"gear/{id}")
            lod.append(next(data))
        return lod

    def get_athlete_stats(self, ids: List[str]) -> List:
        """
        Get Athlets Stats

        :param id: athelete_id (must be same as authenticated athlete)
        :type id: str
        :return: data from api
        :rtype: dict
        """
        lod = []
        for id in ids:
            mds_logger.info("Fetching Athlete Stats Data")
            data = self._get(endpoint=f"athletes/{id}/stats")
            lod.append(next(data))
        return lod

    def get_activities(
        self,
        before_epoch: Optional[int] = None,
        after_epoch: Optional[int] = None,
        per_page: Optional[int] = 100,
    ) -> Generator[List[Dict], None, None]:
        """
        Get activity data from the authorized athlete

        :param before_epoch: Epoch timestamp to retreive data from before that time, defaults to None
        :type before_epoch: Optional[int], optional
        :param after_epoch: Epoch timestamp to retreive data from after that time, defaults to None
        :type after_epoch: Optional[int], optional
        :param per_page: Number of activities to return in call, defaults to None

        :yield: Data from API
        :rtype: Generator[List[Dict], None, None]
        """
        mds_logger.info("Fetching Athelete Activities")

        query_params = {}

        # Format the Query Params
        if per_page:
            query_params["per_page"] = per_page
        if before_epoch:
            query_params["before"] = before_epoch
        if after_epoch:
            query_params["after"] = after_epoch

        data = self._get(endpoint="athlete/activities", query_params=query_params)
        yield from data

    def _get(self, endpoint: str, query_params: Optional[Dict] = None) -> Generator:
        """
        Helper Method for get requests

        :param endpoint: endpoint that gets added to base url
        :type endpoint: str
        :raises Exception: Failed API Call
        :return: API Data returned
        :rtype: dict
        """
        url = f"{self.base_url}{endpoint}"
        mds_logger.info(f"Get Request: {url} - params {query_params}")
        headers = self.get_api_headers()

        response = requests.get(url, headers=headers, params=query_params)

        if response.status_code == 200:
            yield response.json()

        elif response.status_code != 200:
            raise ValueError(
                f"Failed API call with Get request: {response.status_code}\n"
                f"text: {response.text}"
            )



"""
Utilites for miscellaneous reasons
"""
import logging
from hashlib import sha256
from typing import Any


def setup_console_logger(
    logger_name: str = "mds_logger", level: int = logging.DEBUG
) -> logging.Logger:
    """
    Create a python logger

    :param logger_name: logger name to set, defaults to __name__
    :type logger_name: _type_, optional
    :param level: log level, defaults to logging.DEBUG
    :type level: _type_, optional
    :return: the logger
    :rtype: logging.Logger
    """
    logger = logging.getLogger(logger_name)
    if not logger.handlers:
        # Create a logger
        logger = logging.getLogger(logger_name)

        # Set the logging level
        logger.setLevel(level)

        # Create a console handler and set its level
        console_handler = logging.StreamHandler()
        console_handler.setLevel(level)

        # Create a formatter and attach it to the handler
        formatter = logging.Formatter(
            "%(asctime)s - %(levelname)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )
        console_handler.setFormatter(formatter)

        # Add the console handler to the logger
        logger.addHandler(console_handler)

    return logger


def hash_func(value: Any) -> str:
    """
    Hash an input using sha256

    :param value: Input to hash
    :type value: Any
    :return: SHA256 hash
    :rtype: str
    """
    result = sha256(value.encode("utf-8")).hexdigest()
    return result



"""
Date and Time helper utilities
"""
from datetime import date, datetime, timezone
from typing import Optional


def get_current_epoch_time() -> int:
    """
    Gets currents epoch seconds

    Returns:
        int: current epoch time
    """
    return int(datetime.now().timestamp())


def epoch_to_datetime(epoch_time: int) -> datetime:
    """
    Converts Epoch time to Datetime format

    :param epoch_time: epoch seconds
    :type epoch_time: int
    :return: python datetime object
    :rtype: datetime
    """
    return datetime.utcfromtimestamp(epoch_time)


def date_to_epoch(input_date: date, timezone: timezone = timezone.utc) -> int:
    """
    Convert a date to the epoch equivalent

    :param input_date: Input Date
    :type input_date: date
    :param timezone: Timezone to convert to, defaults to timezone.utc
    :type timezone: timezone, optional
    :return: Epoch equivalent of the date
    :rtype: int
    """
    date_time = datetime(
        input_date.year, input_date.month, input_date.day, tzinfo=timezone
    )
    return int(date_time.timestamp())


def datetime_to_epoch(date_time: datetime) -> int:
    """
    Convert a datetime to epoch

    :param date_time: Datetime object
    :type date_time: datetime
    :return: Epoch equivalent
    :rtype: int
    """
    return int(date_time.timestamp())


def get_current_year() -> int:
    """
    Returns the current year

    :return: Current Year
    :rtype: int
    """
    return datetime.now().year


def string_to_datetime(
    str_dt: str,
    time_zone: Optional[timezone] = None,
    format: str = "%Y-%m-%dT%H:%M:%SZ",
) -> datetime:
    """
    Convert a string value to a datetime object

    :param str_dt: Datetime string
    :type str_dt: str
    :param time_zone: timezone to convert the string to
    :type time_zone: Optional[timezone]
    :param format: format to use when parsing, defaults to "YYYY-MM-DDTHH:MM:SSZ"
    :type format: str, optional
    :return: Datetime equivalent
    :rtype: datetime
    """
    if time_zone:
        date_time_obj = datetime.strptime(str_dt, format).replace(tzinfo=time_zone)

    else:
        date_time_obj = datetime.strptime(str_dt, format)

    return date_time_obj


def string_to_date(str_date: str, format: str = "%Y-%m-%d") -> date:
    """
    Converts a string value to a date

    :param str_date: String to convert
    :type str_date: str
    :param format: format to use when parsing, defaults to "%Y-%m-%d"
    :type format: str, optional
    :return: date object
    :rtype: date
    """
    return datetime.strptime(str_date, format).date()



"""
Utilities for working with text (string)
"""
import re


def camel_case_to_snake_case(text: str, to_lower: bool = True) -> str:
    """
    Convert a text from CamelCase to snake_case

    :param text: text to transform
    :type text: str
    :param to_lower: If you want the result to be lowercase, defaults to True
    :type to_lower: bool, optional
    :return: transformed text
    :rtype: str
    """
    regex = "(?<!^)(?=[A-Z])"
    result = re.sub(regex, "_", text)
    # Handle to fix the __{} exception
    result = re.sub(r"^__", "_", result)

    if to_lower:
        return result.lower()
    return result


def remove_special_charachters(
    text: str, regex_remove: str = "[^a-zA-Z0-9_]+", replace_char: str = "_"
) -> str:
    """
    Remove Special Charachters from text and replaces with _

    :param text: text to modify
    :type text: str
    :param regex_remove: regex pattern, defaults to '[^a-zA-Z0-9_]+'
    :type regex_remove: str, optional
    :param replace_char: string to replace it with, defaults to '_'
    :type replace_char: str, optional
    :return: transformed text
    :rtype: str
    """

    pattern = re.compile(regex_remove)

    return pattern.sub(replace_char, text)


def cols_text_to_standard(text: str, upper: bool = True, *args, **kwargs) -> str:
    """
    Convert text to the standard naming convention (for dataframes)

    :param text: text to transform
    :type text: str
    :return: transformed text
    :rtype: str
    """
    text = remove_special_charachters(text, *args, **kwargs)
    text = camel_case_to_snake_case(text, *args, **kwargs)
    if upper:
        text = text.upper()
    return text



"""
Dagster Utilities to help with dagster code
"""
from typing import Union

from dagster import AssetExecutionContext, AssetKey

from core_library.utilities.misc_utils import setup_console_logger

mds_logger = setup_console_logger()


def get_latest_asset_metadata_value(
    context: AssetExecutionContext, asset_name: str, metadata_key: str
) -> str:
    """
    Get the latest metadata result from an asset and the metadata key

    :param context: Dagster asset context
    :type context: AssetExecutionContext
    :param asset_name: The name of the asset
    :type asset_name: str
    :param metadata_key: The name of the metadata key
    :type metadata_key: str
    :return: the value of the metadata key
    :rtype: str
    """
    latest_materialization = context.instance.get_latest_materialization_event(
        AssetKey([asset_name])
    )
    # For typing
    assert latest_materialization is not None
    assert latest_materialization.asset_materialization is not None

    metadata_value = latest_materialization.asset_materialization.metadata[
        metadata_key
    ].text

    return str(metadata_value)


def get_high_watermark_value(
    context: AssetExecutionContext, asset_name: str
) -> Union[str, None]:
    """
    Function to get the metadata 'high_watermark

    :param context: Dagster context
    :type context: AssetExecutionContext
    :param asset_name: Dagster asset name
    :type asset_name: str
    :return: high_watermark value
    :rtype: Union[str, None]
    """
    mds_logger.info("Getting high_watermark from metadata")
    try:
        value = get_latest_asset_metadata_value(
            context, asset_name, metadata_key="high_watermark"
        )
        mds_logger.info(f"high_watermark value = {value}")
        return value
    except KeyError:
        mds_logger.warning("Unable to get high watermark")
        return None



"""
Utilities for working with polars dataframes
"""
from datetime import datetime
from typing import Dict, List, Optional, Union

import polars as pl
from polars.type_aliases import FrameInitTypes

from core_library.utilities.misc_utils import hash_func, setup_console_logger
from core_library.utilities.text_utils import cols_text_to_standard

mds_logger = setup_console_logger()


def pl_create_df(data: FrameInitTypes, schema: Optional[Dict] = None) -> pl.DataFrame:
    """Create a Polars Dataframe

    :param data: Data to create DF
    :type data: FrameInitTypes
    :param schema: Schema dictionary for the polars dataframe
    :type schema: Dict
    :return: Polars Dataframe
    :rtype: pl.DataFrame
    """
    mds_logger.info("Converting to dataframe")
    df = pl.DataFrame(data=data, schema=schema)
    return df


def pl_df_cols_to_standard(df: pl.DataFrame, upper: bool = True):
    """Convert polars dataframe columns to standard used"""
    mds_logger.info("Converting dataframe cols to standard")
    for col in df.columns:
        new_col = cols_text_to_standard(col, upper=upper)
        df = df.rename({col: new_col})

    return df


def pl_concat_dfs(list_of_dfs: List[pl.DataFrame], **kwargs) -> pl.DataFrame:
    """
    Helpful function to take in a list of dataframes and concatenate them

    :param list_of_dfs: List of dataframes
    :type list_of_dfs: List[pl.DataFrame]
    :return: One dataframe
    :rtype: pl.DataFrame
    """

    result_df = pl.concat(items=list_of_dfs, **kwargs)

    return result_df


def pl_aggregate_column(
    df: pl.DataFrame, column_name: str, agg_type: str
) -> Union[int, str, datetime, None]:
    """
    Aggregate a single column and get back one record as a native python type

    :param df: Dataframe to aggregate
    :type df: pl.DataFrame
    :param column_name: Column to aggregate
    :type column_name: str
    :param agg_type: type of aggregation to use
    :type agg_type: str
    :return: Result from dataframe in native python type
    :rtype: Union[int, str, datetime, None]
    """
    result = None

    if df.is_empty():
        mds_logger.info("Empty Dataframe")
        return result

    if agg_type == "max":
        result = df.select(pl.col(column_name)).max()[column_name][0]
    elif agg_type == "min":
        result = df.select(pl.col(column_name)).min()[column_name][0]

    return result


def pl_log_dataframe(
    df: pl.DataFrame,
    n_rows: int = 3,
    order_by: Optional[str] = None,
    ascending: Optional[bool] = None,
    only_cols: Optional[bool] = False,
) -> None:
    """
    Log a dataframe helpful for viewing at different stages

    :param df: Polars dataframe to log
    :type df: pl.DataFrame
    :param n_rows: Number of records to show
    :type n_rows: int
    :param order_by: If you want to order the dataframe, defaults to None
    :type order_by: Optional[str], optional
    :param ascending: Sort order if passing in order_by col, defaults to None
    :type ascending: Optional[bool], optional
    """
    # TODO: Implement this functionality
    if (order_by and not ascending) or (ascending and not order_by):
        Exception("order_by and ascending arguments are both required")

    if only_cols:
        mds_logger.info(f"DF Columns: {df.columns}")
        return None

    mds_logger.info("Dataframe Preview:")
    # TODO: If there are a lot of columns it is weird

    mds_logger.info(df)
    return None


def pl_check_empty_df(df: pl.DataFrame) -> bool:
    """
    Check if a dataframe is empty

    :param df: Input dataframe
    :type df: pl.DataFrame
    :return: True if empty
    :rtype: bool
    """
    if df.is_empty():
        mds_logger.info("Dataframe is emtpy")
        return True
    return False


def pl_concat_str(
    df: pl.DataFrame, cols: Union[List[str], str], alias: str, ignore_nulls: bool = True
) -> pl.DataFrame:
    """
    Concatenate columns together

    :param df: Input Dataframe
    :type df: pl.DataFrame
    :param cols: Columns to concatenate together
    :type cols: Union[List[str], str]
    :param alias: Output column name
    :type alias: str
    :param ignore_nulls: If false nulls will returrn nulls, defaults to True
    :type ignore_nulls: bool, optional
    :return: Output Dataframe
    :rtype: pl.DataFrame
    """
    df = df.with_columns(pl.concat_str(cols, ignore_nulls=ignore_nulls).alias(alias))
    return df


def pl_hash_func(
    df: pl.DataFrame,
    cols: Union[str, List],
    alias: str = "KEY_HASH",
    drop_concat_col: bool = False,
) -> pl.DataFrame:
    """
    Dataframe hash a list of cols - will automatically concat them together if multiple

    :param df: Dataframe to hash
    :type df: pl.DataFrame
    :param cols: Cols to hash
    :type cols: Union[str, List]
    :param alias: New column name, defaults to 'KEY_HASH'
    :type alias: str, optional
    :return: Hashed Dataframe
    :rtype: pl.DataFrame
    """
    df = pl_concat_str(df, cols=cols, alias="KEY_CONCAT")

    df = df.with_columns(pl.col("KEY_CONCAT").map_elements(hash_func).alias(alias))

    if drop_concat_col:
        df = df.drop("KEY_CONCAT")

    return df


def pl_add_standard_cols(df: pl.DataFrame, hash_cols: List[str]) -> pl.DataFrame:
    """
    Adds standard cols to a dataframe on ingestions

    :param df: Dataframe to add cols to
    :type df: pl.DataFrame
    :param hash_cols: What columns to hash
    :type hash_cols: List[str]
    :return: Standard Dataframe
    :rtype: pl.DataFrame
    """
    mds_logger.info("Adding hash cols and ingestion time to dataframe")

    # Add a hash column of the unique keys
    df = pl_hash_func(df, hash_cols)

    # Add an ETL timestamp column
    df = df.with_columns(INGESTION_DATE_TIME=datetime.now())

    return df



"""
Utilities to help with data
"""
from typing import Any, Dict, List, Union


def key_values_in_lod(data: Union[List[Dict], None], select_key: str) -> List[Any]:
    """
    Returns the value of the key from a List of Dictionary

    :param data: LOD with Data
    :type data: List[Dict]
    :param select_key: Key that you want to return
    :type select_key: str
    :return: List of the data
    :rtype: List[Any]
    """
    if data is None:
        return []
    data = [item[select_key] for item in data]
    return data
